import numpy as np

from .base import Transformer


class Rigid(Transformer):
    """Use a rigid transformation function to align the pairs of cells.

    Rigid trasformations are constrained to the operations of rotation,
    reﬂection, translation, and combinations of these.
    """
    def __init__(self):
        super().__init__()

    def fit(self, A, B):
        # See http://nghiaho.com/?page_id=671
        # center
        A_centroid = np.mean(A, axis=0)
        B_centroid = np.mean(B, axis=0)

        H = np.dot((A - A_centroid).T, B - B_centroid)
        U, S, Vt = np.linalg.svd(H)
        R = np.dot(Vt.T, U.T)
        detR = np.linalg.det(R)
        x = np.identity(Vt.T.shape[1])
        x[x.shape[0]-1, x.shape[1]-1] = detR
        R = np.linalg.multi_dot([Vt.T, x, U.T])

        t = B_centroid.T - np.dot(R, A_centroid.T)

        model = {
            'R': R,
            't': t
        }
        return model

    def transform(self, model, A):
        return np.dot(model['R'], A.T).T + model['t']

    def chain(self, model, step_model, step_number):
        # For rigid transforms, since we can learn the final function simply
        # by learning the optimal rotation from the original A to the final A,
        # we don't need to maintain an overall model during fitting. See
        # finalize.
        return None

    def finalize(self, model, A_orig, A_final):
        return self.fit(A_orig, A_final)



import logging

import numpy as np
import torch

from .base import Transformer


class Affine(Transformer):
    """ Use an affine transformation function to align the pairs of cells.

    The affine function is of the form ``f(x) = Wx + b``, where ``W`` and ``b``
    are the learnable weights. ``W`` has the shape (genes, genes) and ``b``
    (the bias term) has shape (genes,).

    Parameters
    ----------
    optim : {'adam', 'sgd'}
        Which torch optimizer to use.

    lr : float
        Learning rate to use in gradient descent.

    epochs : int
        Number of iterations to run gradient descent.
    """
    def __init__(self, optim='adam', lr=1e-3, epochs=1000):
        self.optim = optim
        self.lr = lr
        self.epochs = epochs
        super().__init__()

    def fit(self, A, B):
        log = logging.getLogger(__name__)
        d = A.shape[1]
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        log.info(f'Using device {device}')
        A = torch.from_numpy(A).float().to(device)
        B = torch.from_numpy(B).float().to(device)
        f = torch.nn.Sequential()
        f.add_module('lin', torch.nn.Linear(d, d, bias=True))
        f.to(device)
        if self.optim == 'adam':
            optimizer = torch.optim.Adam(f.parameters(), lr=self.lr)
        else:
            optimizer = torch.optim.SGD(f.parameters(), lr=self.lr,
                                        momentum=0.9, nesterov=True)
        f.train()
        for e in range(self.epochs):
            optimizer.zero_grad()
            loss = torch.mean(torch.norm(f(A) - B, p=2, dim=1)**2)
            if e % 100 == 0:
                log.info(f'\tEpoch: {e}/{self.epochs}, loss: {loss.item()}')
            loss.backward()
            optimizer.step()
        # theta is the augmented matrix wich includes weights W and bias in
        # a single matrix
        theta = np.zeros((d + 1, d + 1))
        theta[:d, :d] = f[0].weight.data.cpu().numpy()
        theta[:d, -1] = f[0].bias.data.cpu().numpy()
        theta[-1, -1] = 1.

        model = {
            'theta': theta,
        }
        return model

    def transform(self, model, A):
        d = A.shape[1]
        W = model['theta'][:d, :d]
        bias = model['theta'][:d, -1]
        return np.dot(W, A.T).T + bias

    def chain(self, model, step_model, step_number):
        # Affine functions can be composed easily when represented in their
        # augmented matrix form, simply left multiply their (augmented)
        # transformation matrices by each other.
        if model is None:
            return step_model
        else:
            model['theta'] = np.dot(step_model['theta'],
                                    model['theta'])
            return model

    def finalize(self, model, A_orig, A_final):
        # Since we've been updating the overal model in the chain function,
        # we don't need to do anything here.
        return model



#!/usr/bin/env python

"""Tests for `scipr` package."""


import unittest
import io
from contextlib import redirect_stdout, redirect_stderr
from pathlib import Path
import tempfile

import numpy as np
import anndata

import scipr
from scipr.matching import Closest
from scipr.transform import Rigid, Affine


class TestScipr(unittest.TestCase):
    """Tests for `scipr` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        np.random.seed(1817)
        self.A = np.random.random((10, 100))
        self.B = np.random.random((20, 100))
        # Example Match and Transform objects,
        # useful to instantiate basic SCIPR objects
        self.match = Closest()
        self.rigid_transform = Rigid()
        self.affine_transform = Affine()

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_l2_normalization(self):
        """Test L2 input normalization."""
        model = scipr.SCIPR(match_algo=self.match,
                            transform_algo=self.rigid_transform,
                            input_normalization='l2')
        normalized = model._apply_input_normalization(self.A)
        norm = np.linalg.norm(normalized, axis=1)
        self.assertTrue(np.allclose(norm, 1.0))

    def test_001_std_normalization(self):
        """Test standard scaling input normalization"""
        model = scipr.SCIPR(match_algo=self.match,
                            transform_algo=self.rigid_transform,
                            input_normalization='std')
        normalized = model._apply_input_normalization(self.A)
        means = np.mean(normalized, axis=0)
        stds = np.std(normalized, axis=0)
        self.assertTrue(np.allclose(means, 0.))
        self.assertTrue(np.allclose(stds, 1.))

    def test_002_log_normalization(self):
        """Test Seurat-style log input normalization"""
        model = scipr.SCIPR(match_algo=self.match,
                            transform_algo=self.rigid_transform,
                            input_normalization='log')
        normalized = model._apply_input_normalization(self.A)
        feature_counts = np.expand_dims(np.sum(self.A, axis=1), 1)
        correct = np.log1p((self.A / feature_counts) * 10000)
        self.assertTrue(np.allclose(normalized, correct))

    def test_003_raise_unfitted(self):
        """Test raises Runtime Exception if transform called before fit"""
        model = scipr.SCIPR(match_algo=self.match,
                            transform_algo=self.rigid_transform,
                            input_normalization='l2')
        with self.assertRaises(RuntimeError):
            model.transform(self.A)

    def test_004_test_transform_same_shape(self):
        """Test tranforming returns data of same shape as input"""
        model = scipr.SCIPR(match_algo=self.match,
                            transform_algo=self.rigid_transform,
                            input_normalization='l2')
        model.fit(self.A, self.B)
        transformed = model.transform(self.A)
        self.assertEqual(self.A.shape, transformed.shape)

    def test_005_test_logging(self):
        """Test fit function emits logging information"""
        with self.assertLogs(level='INFO') as cm:
            model = scipr.SCIPR(match_algo=self.match,
                                transform_algo=self.rigid_transform,
                                input_normalization='l2')
            model.fit(self.A, self.B)
        self.assertEqual(cm.output[:2],
                         ['INFO:scipr.scipr:Applying L2 normalization',
                          'INFO:scipr.scipr:Applying L2 normalization'])

    def test_006_no_stdout_in_normal_conditions(self):
        """Test no output (silent) from SCIPR methods under normal
        conditions (no warnings)"""
        f = io.StringIO()
        with redirect_stdout(f):
            model = scipr.SCIPR(match_algo=self.match,
                                transform_algo=self.affine_transform,
                                input_normalization='l2',
                                n_iter=2)
            model.fit(self.A, self.B)
            model.transform(self.A)
        self.assertEqual(f.getvalue(), '')

    def _does_folder_contain_event_file(self, folder):
        for f in folder.iterdir():
            if 'events.out.tfevents' in f.name:
                return True
        return False

    def test_007_tensorboard_auto_directory(self):
        """Test auto creates directory for event files if not specified"""
        f = io.StringIO()
        with redirect_stderr(f):
            model = scipr.SCIPR(match_algo=self.match,
                                transform_algo=self.rigid_transform,
                                input_normalization='l2',
                                n_iter=2)
            model.fit(self.A, self.B, tensorboard=True)
        stdout = f.getvalue()
        self.assertIn('tensorboard_dir is not specified', stdout)
        tboard_path = Path(stdout.split()[-1])
        self.assertTrue(tboard_path.exists())
        self.assertTrue(self._does_folder_contain_event_file(tboard_path))

    def test_008_tensorboard_custom_directory(self):
        """Test creates specified directory for event files"""
        with tempfile.TemporaryDirectory() as tmp_dir:
            tboard_path = Path(tmp_dir)
            model = scipr.SCIPR(match_algo=self.match,
                                transform_algo=self.rigid_transform,
                                input_normalization='l2',
                                n_iter=2)
            model.fit(self.A, self.B, tensorboard=True,
                      tensorboard_dir=tboard_path)
            self.assertTrue(tboard_path.exists())
            self.assertTrue(self._does_folder_contain_event_file(tboard_path))

    def _test_anndata_helper(self):
        model = scipr.SCIPR(match_algo=self.match,
                            transform_algo=self.affine_transform,
                            n_iter=3,
                            input_normalization='l2')
        adata = anndata.AnnData(X=np.concatenate([self.A, self.B], axis=0),
                                obs={
                                    'batch': (['A'] * self.A.shape[0]) +
                                             (['B'] * self.B.shape[0])
                                },
                                dtype=self.A.dtype)
        model.fit_adata(adata, 'batch', 'A', 'B')
        return adata, model

    def test_009_anndata_transform(self):
        """Test fit and transform with AnnData is equivalent to numpy"""
        adata, model = self._test_anndata_helper()
        transformed_adata, idx = model.transform_adata(adata, 'batch', 'A')
        self.assertEqual(transformed_adata.shape[0], np.sum(idx))
        transformed_np = model.transform(self.A)
        self.assertTrue(np.array_equal(transformed_adata, transformed_np))

    def test_010_anndata_transform_inplace(self):
        """Test inplace transform of AnnData is equivalent to numpy"""
        adata, model = self._test_anndata_helper()
        model.transform_adata(adata, 'batch', 'A', inplace=True)
        transformed_adata = adata[adata.obs['batch'] == 'A'].X
        transformed_np = model.transform(self.A)
        self.assertTrue(np.array_equal(transformed_adata, transformed_np))



# iterative point set registration methods for scRNA-seq alignment
import logging

from sklearn.preprocessing import StandardScaler, normalize
import numpy as np
from scipy import spatial


class SCIPR(object):
    """Single Cell Iterative Point set Registration (SCIPR).

    Alignment of scRNA-seq data batches using an adaptation of the Iterative
    Closest Points (ICP) algorithm. SCIPR's core steps are matching the points
    and learning the transform function, and what strategy to use for each of
    these are specified by the user.

    Parameters
    ----------
    match_algo : scipr.matching.Match
        Which matching strategy to use, an instance of a :class:`Match`.

    transform_algo : scipr.transform.Transformer
        Which transformation strategy to use, an instance of a
        :class:`Transformer`.

    n_iter : int
        Number of steps of SCIPR to run. Each step is a matching phase,
        followed by updating the transformation function.

    input_normalization : {'l2', 'std', 'log'}
        Which input normalization to apply to data before aligning with SCIPR.
         - 'l2' : Scale each cell's count vector to have unit norm (vector
           length).
         - 'std' : Scale each gene to have zero mean and unit variance.
         - 'log' : Apply Seurat-style log normalization to each cell's count
           vector (as in Seurat's ``normalize`` function).
    """
    def __init__(self, match_algo, transform_algo, n_iter=5,
                 input_normalization='l2'):
        self.match_algo = match_algo
        self.transform_algo = transform_algo
        self.n_iter = n_iter
        self.input_normalization = input_normalization
        self.fitted = False

    def _log_hparams(self, tboard):
        tboard.add_hparams(hparam_dict={
            'match_algo': str(self.match_algo),
            'transform_algo': str(self.transform_algo),
            'n_iter': self.n_iter,
            'input_normalization': self.input_normalization},
                           metric_dict={})

    def fit(self, A, B, tensorboard=False, tensorboard_dir=None):
        """Fit the model to align to a reference batch.

        Parameters
        ----------
        A : numpy.ndarray
            The "source" batch of cells to align. Dimensions are
            (cellsA, genes).

        B : numpy.ndarray
            The "target" (or "reference") batch data to align to. ``A`` is
            aligned onto ``B``, where ``B`` is unchanged, and remains a
            stationary "reference". Dimensions are (cellsB, genes).

        tensorboard : bool
            If True, enable tensorboard logging of SCIPR algorithm metrics.

        tensorboard_dir : None or str
            If None, will use an automatically generated folder to store
            tensorboard event files. If specified, will place event files in
            the specified directory (creates it if it doesn't already exist).
        """
        log = logging.getLogger(__name__)
        if tensorboard:
            try:
                from torch.utils.tensorboard import SummaryWriter
            except ImportError:
                raise ImportError('SCIPR Tensorboard logging requires extra ' +
                                  "dependencies, install them via 'pip " +
                                  "install scipr[tensorboard]'")
            tboard = SummaryWriter(log_dir=tensorboard_dir)
            if tensorboard_dir is None:
                log.warning('tensorboard_dir is not specified, using auto ' +
                            f'generated folder: {tboard.get_logdir()}')
            self._log_hparams(tboard)
        A = self._apply_input_normalization(A)
        B = self._apply_input_normalization(B)
        kd_B = spatial.cKDTree(B)
        A_orig = A
        for i in range(self.n_iter):
            a_idx, b_idx, distances = self.match_algo(A, B, kd_B)
            avg_distance = np.mean(distances)
            log.info(f'SCIPR step {i}, n-pairs: {len(a_idx)}, ' +
                     f'avg distance: {avg_distance}')
            if tensorboard:
                tboard.add_scalar('num_pairs', len(a_idx), i)
                tboard.add_scalar('avg_distance', avg_distance, i)
            step_model = self.transform_algo._fit_step(A[a_idx], B[b_idx], i)
            A = self.transform_algo.transform(step_model, A)
        self.transform_algo._finalize(A_orig, A)
        self.fitted = True

    def fit_adata(self, adata, batch_key, source, target, tensorboard=False,
                  tensorboard_dir=None):
        """Fit the model to align to a reference batch, taking AnnData input.

        Parameters
        ----------
        adata : anndata.AnnData
            Annotated data object containing the two "source" and "target" cell
            batches. Dimensions are (cells, genes).

        batch_key : str
            The name of the column in `adata.obs` which contains the batch
            annotations.

        source,target : str
            The batch annotation values of the "source" and "target" batches.

        tensorboard : bool
            If True, enable tensorboard logging of SCIPR algorithm metrics.

        tensorboard_dir : None or str
            If None, will use an automatically generated folder to store
            tensorboard event files. If specified, will place event files in
            the specified directory (creates it if it doesn't already exist).

        See Also
        --------
        fit
        """
        A = adata[adata.obs[batch_key] == source].X
        B = adata[adata.obs[batch_key] == target].X
        self.fit(A, B, tensorboard, tensorboard_dir)

    def transform(self, A):
        """Apply alignment to a batch of cells.

        Cells are transformed to be aligned to the same "reference" batch from
        the :meth:`fit` method.

        Parameters
        ----------
        A : numpy.ndarray
            The batch of cells to align. Dimensions are (cellsA, genes).

        Returns
        -------
        numpy.ndarray
            The aligned batch of cells, same shape as input ``A``.

        Raises
        ------
        RuntimeError
            If this method is called before :meth:`fit` method.
        """
        if not self.fitted:
            raise RuntimeError('Must call "fit" before "transform"!')
        A = self._apply_input_normalization(A)
        return self.transform_algo._transform(A)

    def transform_adata(self, adata, batch_key, batch, inplace=False):
        """Apply alignment to a batch of cells, taking AnnData input.

        Cells are transformed to be aligned to the same "reference" batch from
        the :meth:`fit` method.

        Parameters
        ----------
        adata : anndata.AnnData
            Annotated data object containing the batch of cells to align.
            Dimensions are (cells, genes).

        batch_key : str
            The name of the column in `adata.obs` which contains the batch
            annotations.

        batch : str
            The batch annotation value of the cells to align.

        inplace : bool
            If True, then replace the cells (observations) in adata.X which
            have the `batch` annotation with the new transformed values. Beware
            that these transformed cells will have been first normalized then
            transformed, so they are transformed to align to a *normalized*
            representation (i.e. using the normalizatio specified in the
            constructor). Whereas the other cells in `adata` may not be in this
            normalized representation.

            Otherwise, return the tuple (transformed numpy.ndarray,
            row indexer into adata of the transformed cells).

        Returns
        -------
        transformed : numpy.ndarray
            The aligned batch of cells, same shape as input batch. Only
            provided if `inplace` is True.

        indexer : numpy.ndarray
            Boolean row indexer into `adata` of the transformed cells. Only
            provided if `inplace` is True.

        Raises
        ------
        RuntimeError
            If this method is called before :meth:`fit` method.

        See Also
        --------
        transform
        """
        indexer = adata.obs[batch_key] == batch
        transformed = self.transform(adata[indexer].X)
        if inplace:
            adata.X[indexer, :] = transformed
        else:
            return transformed, indexer

    def _apply_input_normalization(self, X):
        log = logging.getLogger(__name__)
        if self.input_normalization == 'std':
            log.info('Applying Standardization normalization')
            scaler = StandardScaler().fit(X)
            return scaler.transform(X)
        elif self.input_normalization == 'l2':
            log.info('Applying L2 normalization')
            return normalize(X)
        elif self.input_normalization == 'log':
            log.info('Applying log normalization')
            return np.log1p(X / X.sum(axis=1, keepdims=True) * 1e4)



import logging

import torch
import torch.nn as nn

from .base import Transformer

activations = {
    'tanh': nn.Tanh,
    'sigmoid': nn.Sigmoid,
    'relu': nn.ReLU,
    'leaky_relu': nn.LeakyReLU
}


class _Autoencoder(nn.Module):
    def __init__(self, input_size, layer_sizes, act='tanh', dropout=0.0,
                 batch_norm=False, last_layer_act='tanh', tie_weights=True):
        super(_Autoencoder, self).__init__()
        self.input_size = input_size
        self.layer_sizes = layer_sizes
        self.act = act
        self.dropout = dropout
        self.batch_norm = batch_norm
        self.last_layer_act = last_layer_act
        self.tie_weights = tie_weights

        self.encoder = nn.Sequential()
        self.decoder = nn.Sequential()
        prev_size = self.input_size
        for layer, size in enumerate(layer_sizes):
            # Apply dropout
            if self.dropout > 0:
                self.encoder.add_module('enc_dropout_{}'.format(layer),
                                        nn.Dropout(p=self.dropout))
            # Linearity
            self.encoder.add_module('enc_lin_{}'.format(layer),
                                    nn.Linear(prev_size, size))
            # BN
            if self.batch_norm:
                self.encoder.add_module('enc_batch_norm_{}'.format(layer),
                                        nn.BatchNorm1d(size))
            # Finally, non-linearity
            self.encoder.add_module('enc_{}_{}'.format(act, layer),
                                    activations[act]())
            prev_size = size

        reversed_layer_list = list(self.encoder.named_modules())[::-1]
        decode_layer_count = 0
        for name, module in reversed_layer_list:
            if 'lin_' in name:
                size = module.weight.data.size()[1]
                if self.dropout > 0:
                    self.decoder.add_module(
                        'dec_dropout_{}'.format(decode_layer_count),
                        nn.Dropout(p=self.dropout))
                # Linearity
                linearity = nn.Linear(prev_size, size)
                if self.tie_weights:
                    linearity.weight.data = module.weight.data.transpose(0, 1)
                self.decoder.add_module(
                    'dec_lin_{}'.format(decode_layer_count), linearity)
                # if decode_layer_count < len(self.layer_sizes) - 1:
                # if True:
                if not (decode_layer_count == (len(self.layer_sizes) - 1) and
                   self.last_layer_act is None):
                    # BN
                    if self.batch_norm:
                        self.decoder.add_module(
                            'dec_batch_norm_{}'.format(decode_layer_count),
                            nn.BatchNorm1d(size))
                    # Finally, non-linearity
                    self.decoder.add_module(
                        'dec_{}_{}'.format(last_layer_act,
                                           decode_layer_count),
                        activations[last_layer_act]())
                prev_size = size
                decode_layer_count += 1

    def forward(self, x):
        encoded = self.encoder(x)
        reconstructed = self.decoder(encoded)
        return reconstructed


class StackedAutoEncoder(Transformer):
    """Use multiple autoencoders to align the pairs of cells.

    Fit a "stack" of autoencoders, the output of one feeding in as the input
    into the next one (thereby "composing" them). At each step of SCIPR, the
    next autoencoder is fitted, and then added onto the overall stack. Since
    these are autoencoders, the output dimensions of each are the same as the
    input, so the dimensions are maintained.

    Will automatically search for and use a GPU if available.

    Parameters
    ----------
    hidden_sizes : list of int
        The sizes (widths) of the hidden layers of each autoencoder. This is
        one side of the "funnel" of the autoencoder architecture, and the other
        side is built to be symmetric (same as ``hidden_sizes`` but in
        reverse).

    act : {'leaky_relu', 'relu', 'sigmoid', 'tanh'}
        Which non-linear activation function to use for the autoencoders.

    last_layer_act : {None, 'leaky_relu', 'relu', 'sigmoid', 'tanh'}
        Which non-linear activation function to use for the final layer
        (output) of the autoencoders. ``None`` means no non-linearity. See
        Warnings below.

    optim : {'adam', 'sgd'}
        Which torch optimizer to use.

    lr : float
        Learning rate to use in gradient descent.

    epochs : int
        Number of iterations to run gradient descent.

    Warnings
    --------
    Be aware that your choice of input normalization to SCIPR might have
    implications for your choice of the ``last_layer_act`` parameter. For
    exmaple, If your input normalization scales your input features to
    [0, 1.0], then you may want your final layer to use a sigmoid activation.
    Or if your input normalization allows for input features to be (-\\infty,
    +\\infty), then having `None` as the last layer's activation might be best.
    This is just something to keep in mind, ultimately you may choose what
    performs the best alignment for you.
    """
    def __init__(self, hidden_sizes=[64], act='leaky_relu',
                 last_layer_act=None, optim='adam', lr=1e-3, epochs=1000):
        self.hidden_sizes = hidden_sizes
        self.act = act
        self.last_layer_act = last_layer_act
        self.optim = optim
        self.lr = lr
        self.epochs = epochs
        self.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')
        log = logging.getLogger(__name__)
        log.info(f'Using device {self.device}')
        super().__init__()

    def fit(self, A, B):
        log = logging.getLogger(__name__)
        d = A.shape[1]
        A = torch.from_numpy(A).float().to(self.device)
        B = torch.from_numpy(B).float().to(self.device)
        f = _Autoencoder(input_size=d, layer_sizes=self.hidden_sizes,
                         act=self.act, dropout=0.0, batch_norm=False,
                         last_layer_act=self.last_layer_act, tie_weights=True)
        f.to(self.device)
        if self.optim == 'adam':
            optimizer = torch.optim.Adam(f.parameters(), lr=self.lr)
        else:
            optimizer = torch.optim.SGD(f.parameters(), lr=self.lr,
                                        momentum=0.9, nesterov=True)
        f.train()
        for e in range(self.epochs):
            optimizer.zero_grad()
            loss = torch.mean(torch.norm(f(A) - B, p=2, dim=1)**2)
            if e % 100 == 0:
                log.info(f'\tEpoch: {e}/{self.epochs}, loss: {loss.item()}')
            loss.backward()
            optimizer.step()
        model = {
            'autoencoder': f
        }
        return model

    def transform(self, model, A):
        A = torch.from_numpy(A).float().to(self.device)
        model['autoencoder'].eval()
        A = model['autoencoder'].forward(A)
        A = A.detach().cpu().numpy()
        return A

    def chain(self, model, step_model, step_number):
        if model is None:
            model = {'autoencoder': nn.Sequential()}
        model['autoencoder'].add_module(f'autoencoder_{step_number}',
                                        step_model['autoencoder'])
        return model

    def finalize(self, model, A_orig, A_final):
        log = logging.getLogger(__name__)
        log.info('Final model:\n' + str(model))
        return model



from abc import ABC, abstractmethod


class Transformer(ABC):
    """Base class for all transformation function objects.

    Your transformation functions should also subclass this class.
    """
    def __init__(self):
        self._model = None

    def _fit_step(self, A, B, step_number):
        step_model = self.fit(A, B)
        self._model = self.chain(self._model, step_model, step_number)
        return step_model

    @abstractmethod
    def fit(self, A, B):
        """Return a model to transform A onto B.

        Each point in A correspods to the point in B at the same index (as
        chosen by the matching algorithm of SCIPR). The Transformer is fitted
        to learn a function to move points in A closer to their corresponding
        point in B.

        Parameters
        ----------
        A : numpy.ndarray
            The selected "source" cells to align. Dimensions are
            (cells, genes).
        B : numpy.ndarray
            The "target" cells which correspond to each of the "source" cells
            in ``A``. Dimensions are the same as ``A``, so that each row in
            ``B`` is a cell that is paired up with the same row (cell) in
            ``A``.

        Returns
        -------
        model : dict
            The fitted model parameters (state) of the transformation function.
            For example, weights and biases.
        """
        pass

    @abstractmethod
    def transform(self, model, A):
        """Use the given model to transform the data.

        Parameters
        ----------
        model : dict
            The fitted model parameters (state) of the transformation function
            to use to transform ``A``.

        A : numpy.ndarray
            The cells to transform (i.e. to "align"), dimensions are
            (cells, genes).

        Returns
        -------
        numpy.ndarray
            The trasformation of ``A``, same shape as input ``A``.

        """
        pass

    def _transform(self, A):
        return self.transform(self._model, A)

    @abstractmethod
    def chain(self, model, step_model, step_number):
        """Update the overall model.

        Update the transformation function's parameters with the fitted
        parameters of the latest step. The overall alignment function we are
        learning is the composition of the transformation functions learned at
        each step, and you need to provide the logic for how to compose these
        functions here.

        Parameters
        ----------
        model : dict
            The current state of the overall model parameters, before fitting
            the latest step.
        step_model : dict
            The fitted model parameters (state) of the transformation function
            from the latest step.
        step_number : int
            The number of the current step in the SCIPR algorithm.

        Returns
        -------
        model : dict
            The updated overall model that is the fitted weights from all of
            prior steps, composed with the weights from the current latest
            step.
        """
        pass

    @abstractmethod
    def finalize(self, model, A_orig, A_final):
        """Finalize the overall model at the end of SCIPR.

        If there are any final operations necessary to fit the overall model,
        take them here.

        Parameters
        ----------
        model : dict
            The current state of the overall model parameters, after fitting
            and updating at all of the steps of SCIPR.
        A_orig : numpy.ndarray
            The original "source" batch before the first step of SCIPR, which
            we are fitting to align.
        A_final : numpy.ndarray
            The final state of the "source" batch after the last step of SCIPR,
            the result of transforming ``A_orig`` at each step.

        Returns
        -------
        model : dict
            The final overall model that is the fitted weights from all of the
            steps of the SCIPR algorithm.
        """
        pass

    def _finalize(self, A_orig, A_final):
        self._model = self.finalize(self._model, A_orig, A_final)



from abc import ABC, abstractmethod
import math
import datetime
from collections import defaultdict
import logging

from scipy import spatial
from scipy.optimize import linear_sum_assignment
import numpy as np


class Match(ABC):
    """Base class for all matching function objects.

    Your matching functions should also subclass this class.
    """
    @abstractmethod
    def match(self, A, B, kd_tree_B):
        """ Find matching pairs of cells between two batches.

        Parameters
        ----------
        A : numpy.ndarray
            The "source" batch of cells to align. Dimensions are
            (cellsA, genes).

        B : numpy.ndarray
            The "target" (or "reference") batch data to align to. Dimensions
            are (cellsB, genes).

        kd_tree_B : scipy.spatial.ckdtree.cKDTree
            A KD tree of the ``B`` batch for fast queries, since ``B`` is the
            stationary "reference" batch which does not move, SCIPR computes
            once at the beginning and passes it along to the matching algorithm
            at each step.

        Returns
        -------
        A_indices : numpy.ndarray
            Index array into ``A``, selecting the matched cells in ``A``. Of
            length S.

        B_indices : numpy.ndarray
            Index array into ``B``, selecting the matched cells in ``B``. Also
            of length S. Each element in this array corresponds to its match in
            ``A_indices``.

        distances : numpy.ndarray
            The distances between the pairs in ``A_indices`` and ``B_indices``,
            also of length S.
        """
        pass

    def __call__(self, A, B, kd_tree_B):
        return self.match(A, B, kd_tree_B)


class Closest(Match):
    """Use the classic "closest" strategy to assign pairs.

    For each cell in source batch ``A``, pair it with the closest cell to it in
    the target batch ``B``.
    """
    # TODO: add 'prune_matches' option in a constructor here

    def match(self, A, B, kd_tree_B):
        log = logging.getLogger(__name__)
        t0 = datetime.datetime.now()
        distances, B_indices = kd_tree_B.query(A)
        A_indices = np.arange(A.shape[0])
        t1 = datetime.datetime.now()
        time_str = _pretty_tdelta(t1 - t0)
        log.info('Closest matching took ' + time_str)
        return A_indices, B_indices, distances


class Hungarian(Match):
    """Use the Hungarian algorithm for the assignment problem to assign pairs.

    The "Hungarian" method is an efficient algorithm to solve the assignment
    problem, where finding pairs between two sets ``A`` and ``B`` is treated
    as a bipartite matching problem.

    Parameters
    ----------
    frac_to_match : float
        If not 1.0, then this is the fraction of the matches to keep. All of
        the matches are sorted from smallest distance to largest, and then only
        the top ``n`` are returned, where ``n`` is the ``frac_to_match``
        portion of the smaller of the two sets of cells.
    """
    def __init__(self, frac_to_match=1.0):
        self.frac_to_match = frac_to_match

    def match(self, A, B, kd_tree_B):
        log = logging.getLogger(__name__)
        t0 = datetime.datetime.now()
        dist_mat = spatial.distance.cdist(A, B)
        A_indices, B_indices = linear_sum_assignment(dist_mat)
        t1 = datetime.datetime.now()
        time_str = _pretty_tdelta(t1 - t0)
        log.info('Hungarian matching took ' + time_str)
        distances = dist_mat[A_indices, B_indices]
        if self.frac_to_match < 1.0:
            n_to_match = math.floor(self.frac_to_match * min(A.shape[0],
                                                             B.shape[0]))
            idx = np.argsort(distances)
            distances = distances[idx][:n_to_match]
            A_indices = A_indices[idx][:n_to_match]
            B_indices = B_indices[idx][:n_to_match]
        assert(len(np.unique(A_indices)) == len(A_indices))
        assert(len(np.unique(B_indices)) == len(B_indices))
        return A_indices, B_indices, distances


class Greedy(Match):
    """Use the greedy matching algorithm from the SCIPR paper.

    First all pairings between the two sets are sorted by distance from
    smallest to largest, and then selecting pairings proceeds down the list.
    The selection of pairs stops when we have assigned ``alpha`` fraction of
    the source set to pairs. When we are considering a pair, if the ``target``
    point in that pair has already participated in ``beta`` pairs, we do not
    pick it.

    Parameters
    ----------
    alpha : float
        The ``alpha`` hyperparameter in the above algorithm.

    beta : int
        The ``beta`` hyperparameter in the above algorithm.
    """
    def __init__(self, alpha=0.5, beta=2):
        self.alpha = alpha
        self.beta = beta

    def match(self, A, B, kd_tree_B):
        log = logging.getLogger(__name__)
        t0 = datetime.datetime.now()
        dist_mat = spatial.distance.cdist(A, B)
        # Select pairs that should be matched between set A and B,
        # iteratively building up a mask that selects those matches
        mask = np.zeros(dist_mat.shape, dtype=np.float32)
        # sort the distances by smallest->largest
        sorted_idx = np.stack(np.unravel_index(np.argsort(dist_mat.ravel()),
                                               dist_mat.shape), axis=1)
        target_matched_counts = defaultdict(int)
        source_matched = set()
        for i in range(sorted_idx.shape[0]):
            # A tuple, match_idx[0] is index of the pair in set A,
            # match_idx[1] " " B
            match_idx = sorted_idx[i]
            if target_matched_counts[match_idx[1]] < self.beta and \
               match_idx[0] not in source_matched:
                # if the target point in this pair hasn't been matched to too
                # much, and the source point in this pair has never been
                # matched to, then select this pair
                mask[match_idx[0], match_idx[1]] = 1
                target_matched_counts[match_idx[1]] += 1
                source_matched.add(match_idx[0])
            if len(source_matched) > self.alpha * dist_mat.shape[0]:
                # if matched enough of the source set, then stop
                break
        A_indices, B_indices = np.where(mask == 1)
        distances = dist_mat[A_indices, B_indices]
        t1 = datetime.datetime.now()
        time_str = _pretty_tdelta(t1 - t0)
        log.info('Greedy matching took ' + time_str)
        return A_indices, B_indices, distances


class MNN(Match):
    """Use the Mutual Nearest Neighbors strategy to assign pairs.

    For any given cell ``a`` in a set ``A``, if a cell ``b`` in a set ``B`` is
    in the set of nearest neighbors of ``a`` among ``B``, and ``a`` is in the
    set of nearest neighbors of ``b`` among A, then pair ``(a, b)`` is added to
    the set of pairs.

    Parameters
    ----------
    k : int
        The number of neighbors of each cell to consider when finding mutual
        nearest neighbors.
    """
    def __init__(self, k=10):
        self.k = k

    def match(self, A, B, kd_tree_B):
        log = logging.getLogger(__name__)
        t0 = datetime.datetime.now()
        kd_A = spatial.cKDTree(A)

        distances_a2b, B_nb_indices = kd_tree_B.query(A, self.k)
        distances_b2a, A_nb_indices = kd_A.query(B, self.k)
        A_indices, B_indices, distances = [], [], []
        for a_idx, b_neighbors in enumerate(B_nb_indices):
            for b_idx, a_neighbors in enumerate(A_nb_indices):
                if a_idx in a_neighbors and b_idx in b_neighbors:
                    A_indices.append(a_idx)
                    B_indices.append(b_idx)
                    distances.append(kd_tree_B.query(A[a_idx])[0])

        t1 = datetime.datetime.now()
        time_str = _pretty_tdelta(t1 - t0)
        log.info('MNN matching took ' + time_str)
        return np.array(A_indices), np.array(B_indices), np.array(distances)


def _pretty_tdelta(tdelta):
    hours, rem = divmod(tdelta.seconds, 3600)
    mins, secs = divmod(rem, 60)
    return '{:02d}:{:02d}:{:02d}'.format(hours, mins, secs)
